import pydantic

from thanosql._service import ThanoSQLService


class BaseModel(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(
        arbitrary_types_allowed=True, json_encoders={ThanoSQLService: lambda v: str(v)}
    )

    def __repr__(self):
        return f"{self.__class__.__name__}({self.model_dump_json(indent=4)})"



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from tests.faker import fake
from thanosql._error import ThanoSQLNotFoundError, ThanoSQLValueError

if TYPE_CHECKING:
    from thanosql._client import ThanoSQL

column_name = "name"
int_column_name = "price"
# ideally we should delete this dir after testing
# but as there is no API for that, it is somewhat difficult and risky
# so for now we will have a dedicated, fixed folder name for all tests
dir_name = "test"
file_name = "file_image.jpeg"
default_file_path = f"drive/{file_name}"


def test_upload_file(client: ThanoSQL, basic_table_name: str):
    # dir not subpath of drive/
    with pytest.raises(ThanoSQLValueError):
        client.file.upload(path=file_name, dir="..")

    # nonexistent table
    with pytest.raises(ThanoSQLNotFoundError):
        client.file.upload(
            path=file_name,
            db_commit=True,
            table=fake.unique.pystr(8),
            column=column_name,
            dir=dir_name,
        )

    # nonexistent column
    with pytest.raises(ThanoSQLNotFoundError):
        client.file.upload(
            path=file_name,
            db_commit=True,
            table=basic_table_name,
            column=fake.unique.pystr(8),
            dir=dir_name,
        )

    # trying to insert text into integer column
    with pytest.raises(ThanoSQLValueError):
        client.file.upload(
            path=file_name,
            db_commit=True,
            table=basic_table_name,
            column=int_column_name,
            dir=dir_name,
        )

    # upload without db_commit and dir
    # note that we cannot check the existence of uploaded files as it requires user_data_root
    # which can technically be saved, but it increases complexity and safety risk
    res = client.file.upload(path=file_name)
    assert res["data"]["file_path"] == default_file_path

    # upload with db_commit and dir
    target_table = client.table.get(name=basic_table_name)
    record_count_before = target_table.get_records().total

    res = client.file.upload(
        path=file_name,
        db_commit=True,
        table=basic_table_name,
        column=column_name,
        dir=dir_name,
    )
    assert res["data"]["file_path"] == f"drive/{dir_name}/{file_name}"
    assert res["data"]["table_name"] == basic_table_name
    assert res["data"]["column_name"] == column_name

    record_count_after = target_table.get_records().total
    assert record_count_after == record_count_before + 1


def test_get_files(client: ThanoSQL):
    # path must start with drive/
    with pytest.raises(ThanoSQLValueError):
        client.file.list(path=dir_name)

    res = client.file.list(path=f"drive/{dir_name}/*")
    # at least the file we just uploaded
    assert len(res["data"]["matched_pathnames"]) >= 1


def test_delete_file(client: ThanoSQL, basic_table_name: str):
    # path must start with drive/
    with pytest.raises(ThanoSQLValueError):
        client.file.delete(path=dir_name)

    with pytest.raises(ThanoSQLNotFoundError):
        client.file.delete(path=f"drive/{fake.unique.pystr(8)}")

    # delete file inside default drive dir
    res = client.file.delete(path=default_file_path)
    assert "message" in res

    # delete file inside custom dir
    path = f"drive/{dir_name}/{file_name}"

    with pytest.raises(ThanoSQLNotFoundError):
        client.file.delete(
            path=path, db_commit=True, table=fake.unique.pystr(8), column=column_name
        )

    with pytest.raises(ThanoSQLNotFoundError):
        client.file.delete(
            path=path,
            db_commit=True,
            table=basic_table_name,
            column=fake.unique.pystr(8),
        )

    target_table = client.table.get(name=basic_table_name)
    record_count_before = target_table.get_records().total

    res = client.file.delete(
        path=path, db_commit=True, table=basic_table_name, column=column_name
    )
    assert "message" in res

    record_count_after = target_table.get_records().total
    assert record_count_after == record_count_before - 1



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from pydantic import Field, TypeAdapter

from thanosql._service import ThanoSQLService
from thanosql.resources import Column
from thanosql.resources._model import BaseModel

if TYPE_CHECKING:
    from thanosql._client import ThanoSQL


class View(BaseModel):
    name: str
    table_schema: Optional[str] = Field(alias="schema", default=None)
    columns: List[Column] = []
    definition: str = ""


class ViewService(ThanoSQLService):
    def __init__(self, client: ThanoSQL) -> None:
        super().__init__(client=client, tag="view")

    def list(
        self,
        schema: Optional[str] = None,
        verbose: Optional[bool] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> List[View]:
        path = f"/{self.tag}/"
        query_params = self._create_input_dict(
            schema=schema,
            verbose=verbose,
            offset=offset,
            limit=limit,
        )

        raw_response = self.client._request(
            method="get", path=path, query_params=query_params
        )

        views_adapter = TypeAdapter(List[View])
        parsed_response = views_adapter.validate_python(raw_response["views"])
        return parsed_response

    def get(self, name: str, schema: Optional[str] = None) -> View:
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(schema=schema)

        raw_response = self.client._request(
            method="get", path=path, query_params=query_params
        )

        view_adapter = TypeAdapter(View)
        parsed_response = view_adapter.validate_python(raw_response["view"])
        return parsed_response

    def delete(self, name: str, schema: Optional[str] = None) -> dict:
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(schema=schema)

        return self.client._request(
            method="delete", path=path, query_params=query_params
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from thanosql._service import ThanoSQLService

if TYPE_CHECKING:
    from thanosql._client import ThanoSQL


class SchemaService(ThanoSQLService):
    def __init__(self, client: ThanoSQL) -> None:
        super().__init__(client=client, tag="schema")

    def list(self) -> dict:
        path = f"/{self.tag}/"

        return self.client._request(method="get", path=path)

    def create(self, name: str) -> dict:
        path = f"/{self.tag}/{name}"

        return self.client._request(method="post", path=path)



from __future__ import annotations

from ._file import FileService
from ._query import QueryLog, QueryService, QueryTemplate
from ._record import Records
from ._schema import SchemaService
from ._table import (
    BaseColumn,
    BaseTable,
    Column,
    Constraints,
    ForeignKey,
    PrimaryKey,
    Table,
    TableObject,
    TableService,
    TableTemplate,
    Unique,
)
from ._view import View, ViewService

__all__ = [
    "QueryService",
    "TableService",
    "ViewService",
    "SchemaService",
    "FileService",
]



from typing import Any, List

import pandas as pd
from pydantic import model_validator

from thanosql.resources._model import BaseModel


class Records(BaseModel):
    data: List[dict] = []
    total: int = 0

    @model_validator(mode="before")
    @classmethod
    def convert_records_list_only(cls, data: Any) -> Any:
        if isinstance(data, list):
            return {"data": data, "total": len(data)}
        return data

    def to_df(self, **kwargs) -> pd.DataFrame:
        return pd.DataFrame.from_records(self.data, **kwargs)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from tests.faker import fake
from thanosql._error import ThanoSQLNotFoundError
from thanosql.resources import View

if TYPE_CHECKING:
    from _pytest.fixtures import FixtureRequest

    from thanosql._client import ThanoSQL


def test_get_views_success(client: ThanoSQL):
    res = client.view.list()
    assert isinstance(res, list)


def test_get_views_nonexistent_schema(client: ThanoSQL):
    with pytest.raises(ThanoSQLNotFoundError):
        client.view.list(schema=fake.unique.pystr(8))


def test_get_views_limit(client: ThanoSQL, empty_view: dict, basic_view: dict):
    res = client.view.list(limit=2)
    assert isinstance(res, list)

    # at least empty_view and basic_view, at most 2 because of the limit
    assert len(res) == 2


@pytest.mark.parametrize("name", ["empty_view_name", "basic_view_name"])
def test_get_view_success(client: ThanoSQL, name: str, request: FixtureRequest):
    name = request.getfixturevalue(name)
    res = client.view.get(name=name)
    assert isinstance(res, View)
    assert res.name == name


def test_get_view_nonexistent(client: ThanoSQL):
    with pytest.raises(ThanoSQLNotFoundError):
        client.view.get(name=fake.unique.pystr(10))


@pytest.mark.parametrize(
    "view_name",
    ["empty_view_name", "basic_view_name"],
)
def test_delete_view_success(client: ThanoSQL, view_name: str, request: FixtureRequest):
    view_name = request.getfixturevalue(view_name)
    res = client.view.delete(name=view_name)
    assert isinstance(res, dict)

    with pytest.raises(ThanoSQLNotFoundError):
        client.view.get(name=view_name)



from __future__ import annotations

from typing import TYPE_CHECKING

import pandas as pd
import pytest

from tests.faker import fake
from thanosql._error import (
    ThanoSQLAlreadyExistsError,
    ThanoSQLNotFoundError,
    ThanoSQLValueError,
)
from thanosql.resources import QueryLog, QueryTemplate, Records

if TYPE_CHECKING:
    from thanosql._client import ThanoSQL

plain_query_template_name = (
    f"test_plain_{fake.unique.pystr(min_chars=8, max_chars=8).lower()}"
)
basic_query_template_name = (
    f"test_basic_{fake.unique.pystr(min_chars=8, max_chars=8).lower()}"
)
changed_query_template_name = (
    f"test_changed_{fake.unique.pystr(min_chars=8, max_chars=8).lower()}"
)

plain_query_template_string = "LIST THANOSQL MODEL"
basic_query_template_string = "SELECT * FROM {{ table_name }}"
changed_query_template_string = "SELECT {% for col in columns %}{{ col }}{% if not loop.last %}, {% endif %}{% endfor %} FROM {{ table_name }} LIMIT 0"
invalid_query_template_string = "DELETE MODEL {% model_name %}"

query_test_table_name = (
    f"test_query_{fake.unique.pystr(min_chars=8, max_chars=8).lower()}"
)
query_test_selected_columns = ["name", "price"]


def test_create_query_template_invalid(client: ThanoSQL):
    # name too long
    with pytest.raises(ThanoSQLValueError):
        client.query.template.create(name=fake.unique.pystr(min_chars=35, max_chars=40))

    # invalid jinja template
    with pytest.raises(ThanoSQLValueError):
        client.query.template.create(query=invalid_query_template_string)


def test_create_query_template_success(client: ThanoSQL):
    # execute with dry run without any name or query (empty request)
    # check whether auto-naming is working
    res = client.query.template.create(dry_run=True)
    assert isinstance(res, QueryTemplate)
    assert res.query == ""
    assert len(res.parameters) == 0

    temp_name = res.name
    assert temp_name.startswith("query_template_")

    # make sure it is not saved in db
    with pytest.raises(ThanoSQLNotFoundError):
        client.query.template.get(name=temp_name)

    # save to db
    res = client.query.template.create(
        name=plain_query_template_name, query=plain_query_template_string
    )
    assert isinstance(res, QueryTemplate)
    assert res.id != 0
    assert res.name == plain_query_template_name
    assert res.query == plain_query_template_string
    assert len(res.parameters) == 0

    # no need to do thorough checks as we already did in the previous template
    res = client.query.template.create(
        name=basic_query_template_name, query=basic_query_template_string
    )
    assert res.query == basic_query_template_string
    assert res.parameters == ["table_name"]

    # make sure we cannot create templates with the same name again
    with pytest.raises(ThanoSQLAlreadyExistsError):
        client.query.template.create(
            name=plain_query_template_name, query=plain_query_template_string
        )

    with pytest.raises(ThanoSQLAlreadyExistsError):
        client.query.template.create(
            name=basic_query_template_name, query=basic_query_template_string
        )


def test_get_query_template_not_found(client: ThanoSQL):
    with pytest.raises(ThanoSQLNotFoundError):
        client.query.template.get(name=fake.unique.pystr(min_chars=8, max_chars=10))


@pytest.mark.parametrize("name", [plain_query_template_name, basic_query_template_name])
def test_get_query_template_success(client: ThanoSQL, name: str):
    # test whether get is working and whether the template creation is successful
    res = client.query.template.get(name=name)
    assert isinstance(res, QueryTemplate)


def test_get_query_templates_default(client: ThanoSQL):
    res = client.query.template.list()
    assert isinstance(res, list)
    # at least the two templates we just created
    assert len(res) >= 2
    assert isinstance(res[0], QueryTemplate)


def test_update_query_template_invalid(client: ThanoSQL):
    # not found
    with pytest.raises(ThanoSQLNotFoundError):
        client.query.template.update(
            current_name=fake.unique.pystr(min_chars=8, max_chars=10)
        )

    # new name too long
    with pytest.raises(ThanoSQLValueError):
        client.query.template.update(
            current_name=basic_query_template_name,
            new_name=fake.unique.pystr(min_chars=35, max_chars=40),
        )

    # new name already used
    with pytest.raises(ThanoSQLAlreadyExistsError):
        client.query.template.update(
            current_name=basic_query_template_name, new_name=plain_query_template_name
        )

    # invalid jinja template
    with pytest.raises(ThanoSQLValueError):
        client.query.template.update(
            current_name=basic_query_template_name, query=invalid_query_template_string
        )


def test_update_query_template_success(client: ThanoSQL):
    # update name only
    # make sure name is changed but the rest stays the same
    res = client.query.template.update(
        current_name=plain_query_template_name, new_name=changed_query_template_name
    )
    assert res.name == changed_query_template_name
    assert res.query == plain_query_template_string
    assert len(res.parameters) == 0

    # make sure the new name is callable while the old name is not
    with pytest.raises(ThanoSQLNotFoundError):
        client.query.template.get(name=plain_query_template_name)

    res = client.query.template.get(name=changed_query_template_name)
    assert isinstance(res, QueryTemplate)

    # update query only
    # make sure the name is not changed but query and parameters got changed
    res = client.query.template.update(
        current_name=basic_query_template_name, query=changed_query_template_string
    )
    assert res.name == basic_query_template_name
    assert res.query == changed_query_template_string
    assert set(res.parameters) == {"columns", "table_name"}


def test_get_query_logs_default(client: ThanoSQL):
    res = client.query.log.list()
    # at least we did 2 queries in previous tests
    assert len(res["query_logs"]) >= 2
    assert res["total"] >= 2


def test_post_query_invalid(client: ThanoSQL, basic_table_name: str):
    # schema does not exist
    with pytest.raises(ThanoSQLNotFoundError):
        client.query.execute(
            schema=fake.unique.pystr(8),
            table_name=fake.unique.pystr(8),
            query=plain_query_template_string,
        )

    # table already exists
    with pytest.raises(ThanoSQLAlreadyExistsError):
        client.query.execute(
            table_name=basic_table_name, query=plain_query_template_string
        )

    # invalid template input directly
    with pytest.raises(ThanoSQLValueError):
        params = {"model_name": "random_model"}
        client.query.execute(query=invalid_query_template_string, parameters=params)

    # invalid template name
    with pytest.raises(ThanoSQLNotFoundError):
        client.query.execute(template_name=fake.unique.pystr(8))

    # invalid template id -> we most probably will not reach this number
    # but in case someday there will be this many query templates,
    # the invalid id needs to be adjusted
    with pytest.raises(ThanoSQLNotFoundError):
        client.query.execute(template_id=99999)

    # invalid parameter dict
    with pytest.raises(ThanoSQLValueError):
        params = {"table_name": "random_table", "cols": ["column_1", "column_2"]}
        client.query.execute(template_name=basic_query_template_name, parameters=params)


def test_post_query_success(client: ThanoSQL, basic_table_name, empty_table_name):
    params = {"table_name": basic_table_name, "columns": query_test_selected_columns}
    completed_changed_query = f"SELECT name, price FROM {basic_table_name} LIMIT 0"

    # direct entry template
    res = client.query.execute(
        query=changed_query_template_string, parameters=params, max_results=10
    )
    assert isinstance(res, QueryLog)
    assert res.error_result is None
    assert res.destination_schema == "qm"
    assert res.query == completed_changed_query

    # there should be empty record([]) even if max_results is > 0 as we set LIMIT 0 in the query
    assert len(res.records.data) == 0

    # make sure the qm table is created
    qm_table_name = res.destination_table_name
    res = client.table.get(name=qm_table_name, schema="qm")
    assert res

    # direct entry without template or parameters
    res = client.query.execute(
        query=plain_query_template_string, table_name=query_test_table_name
    )
    assert isinstance(res, QueryLog)
    assert res.error_result is None
    assert res.destination_schema == "public"
    assert res.query == plain_query_template_string
    assert res.destination_table_name == query_test_table_name

    # since max_results is 0 by default, there should be no records at all
    assert res.records is None

    # make sure the table is created in public schema
    res = client.table.get(name=query_test_table_name)
    assert res

    # using saved query template
    res = client.query.execute(
        template_name=basic_query_template_name,
        parameters=params,
        table_name=empty_table_name,
        overwrite=True,
    )
    assert isinstance(res, QueryLog)
    assert res.error_result is None
    assert res.destination_schema == "public"
    assert res.query == completed_changed_query
    assert res.destination_table_name == empty_table_name


def test_query_log_records(client: ThanoSQL, empty_table):
    # we want to check if the records section of QueryLog is working properly
    # since we didn't create any tables in this module, we will use the
    # empty_table fixture to ensure that there is at least one table in
    # the list of public tables in information_schema
    query = """
    SELECT table_name FROM information_schema.tables
    WHERE table_schema = 'public'
    """
    res = client.query.execute(query=query, max_results=100)
    assert isinstance(res.records, Records)

    # check if to_df is working and check that records is nonempty
    df = res.records.to_df()
    assert isinstance(df, pd.DataFrame)
    assert len(df.index) > 0


@pytest.mark.parametrize(
    "name", [changed_query_template_name, basic_query_template_name]
)
def test_delete_query_template_success(client: ThanoSQL, name: str):
    res = client.query.template.delete(name=name)
    assert "message" in res

    # make sure the query template is no longer "gettable"
    # also to check whether not found error is raised properly
    with pytest.raises(ThanoSQLNotFoundError):
        client.query.template.get(name=name)



from __future__ import annotations

import os
from typing import Optional

from thanosql._base_client import ThanoSQLBaseClient
from thanosql._error import ThanoSQLValueError
from thanosql.resources import (
    FileService,
    QueryService,
    SchemaService,
    TableService,
    ViewService,
)


class ThanoSQL(ThanoSQLBaseClient):
    def __init__(
        self,
        api_token: Optional[str] = None,
        engine_url: Optional[str] = None,
        api_version: str = "v1",
    ) -> None:
        if api_token is None:
            api_token = os.environ.get("THANOSQL_API_TOKEN", "")
        if not api_token:
            raise ThanoSQLValueError(
                "Please input a valid API token. You can do this either by passing it as a parameter or setting the THANOSQL_API_TOKEN environment variable."
            )

        if engine_url is None:
            engine_url = os.environ.get("THANOSQL_ENGINE_URL", "")
        if not engine_url:
            raise ThanoSQLValueError(
                "Please input a valid engine URL. You can do this either by passing it as a parameter or setting the THANOSQL_ENGINE_URL environment variable."
            )

        super().__init__(token=api_token, base_url=engine_url, version=api_version)

    @property
    def query(self) -> QueryService:
        """Access the QueryService."""
        return QueryService(self)

    @property
    def file(self) -> FileService:
        """Access the FileService."""
        return FileService(self)

    @property
    def schema(self) -> SchemaService:
        """Access the SchemaService."""
        return SchemaService(self)

    @property
    def table(self) -> TableService:
        """Access the TableService."""
        return TableService(self)

    @property
    def view(self) -> ViewService:
        """Access the ViewService."""
        return ViewService(self)



from __future__ import annotations

from thanosql._client import ThanoSQL
from thanosql._version import __version__

__all__ = ["ThanoSQL"]


# In order to actually use these magics, you must register them with a
# running IPython.
def load_ipython_extension(ipython):
    # Load Ipython Magic
    from .magic.magic import ThanosMagic

    """Load the extension in IPython."""
    ipython.register_magics(ThanosMagic)



from __future__ import annotations

import enum
import json
import os
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Union

import pandas as pd
from pydantic import Field, TypeAdapter

from thanosql._error import ThanoSQLValueError
from thanosql._service import ThanoSQLService
from thanosql.resources._model import BaseModel
from thanosql.resources._record import Records

if TYPE_CHECKING:
    from thanosql._client import ThanoSQL


class BaseColumn(BaseModel):
    default: Optional[str] = None
    is_nullable: Optional[bool] = True
    type: str
    name: str


class Column(BaseModel):
    id: Optional[int] = None
    default: Optional[str] = None
    is_nullable: Optional[bool] = True
    type: str
    name: str


class Unique(BaseModel):
    name: Optional[str] = None
    columns: Optional[List[str]] = []


class PrimaryKey(BaseModel):
    name: Optional[str] = None
    columns: Optional[List[str]] = []


class ForeignKey(BaseModel):
    name: Optional[str] = None
    reference_schema: str = "public"
    reference_column: str
    reference_table: str
    column: str


class Constraints(BaseModel):
    unique: Optional[List[Unique]] = None
    primary_key: Optional[PrimaryKey] = None
    foreign_keys: Optional[List[ForeignKey]] = None


class BaseTable(BaseModel):
    name: Optional[str] = None
    table_schema: Optional[str] = Field(alias="schema", default=None)
    columns: Optional[List[BaseColumn]] = None
    constraints: Optional[Constraints] = None


class TableObject(BaseModel):
    columns: Optional[List[BaseColumn]] = None
    constraints: Optional[Constraints] = None


class IfExists(enum.Enum):
    FAIL = "fail"
    APPEND = "append"
    REPLACE = "replace"


class TableService(ThanoSQLService):
    def __init__(self, client: ThanoSQL) -> None:
        super().__init__(client=client, tag="table")

        self.template: TableTemplateService = TableTemplateService(client)

    def _parse_table_response(self, raw_response: dict) -> Table:
        table_adapter = TypeAdapter(Table)
        parsed_response = table_adapter.validate_python(raw_response["table"])
        parsed_response.service = self
        return parsed_response

    def list(
        self,
        schema: Optional[str] = None,
        verbose: Optional[bool] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> List[Table]:
        path = f"/{self.tag}/"
        query_params = self._create_input_dict(
            schema=schema, verbose=verbose, offset=offset, limit=limit
        )

        raw_response = self.client._request(
            method="get", path=path, query_params=query_params
        )

        tables_adapter = TypeAdapter(List[Table])
        parsed_response = tables_adapter.validate_python(raw_response["tables"])
        for table in parsed_response:
            table.service = self
        return parsed_response

    def get(self, name: str, schema: Optional[str] = None) -> Union[Table, dict]:
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(schema=schema)

        raw_response = self.client._request(
            method="get", path=path, query_params=query_params
        )

        return self._parse_table_response(raw_response)

    def update(
        self, name: str, schema: Optional[str] = None, table: Optional[BaseTable] = None
    ) -> Table:
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(schema=schema)
        payload = self._create_input_dict(table=table)

        raw_response = self.client._request(
            method="put", path=path, query_params=query_params, payload=payload
        )

        return self._parse_table_response(raw_response)

    def create(
        self,
        name: str,
        schema: Optional[str] = None,
        table: Optional[TableObject] = None,
    ) -> Table:
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(schema=schema)
        payload = self._create_input_dict(table=table)

        raw_response = self.client._request(
            method="post", path=path, query_params=query_params, payload=payload
        )

        return self._parse_table_response(raw_response)

    def upload(
        self,
        name: str,
        file: Optional[Union[str, os.PathLike]] = None,
        df: Optional[pd.DataFrame] = None,
        schema: Optional[str] = None,
        table: Optional[TableObject] = None,
        if_exists: str = "fail",
    ) -> Table:
        try:
            if_exists_enum = IfExists(if_exists)
        except Exception as e:
            raise ThanoSQLValueError(str(e))

        if file and df is not None:
            raise ThanoSQLValueError(
                "Cannot use both file and DataFrame for upload at the same time."
            )

        if file:
            path = f"/{self.tag}/{name}/upload/"

            file_extension = Path(file).suffix.lower()
            if file_extension == ".csv":
                path = path + "csv"
            elif file_extension in [
                ".xls",
                ".xlsx",
                ".xlsm",
                ".xlsb",
                ".odf",
                ".ods",
                ".odt",
            ]:
                path = path + "excel"
            else:
                raise ThanoSQLValueError(
                    "Invalid format: only CSV and Excel files possible."
                )

            query_params = self._create_input_dict(
                schema=schema, if_exists=if_exists_enum.value
            )
            payload = self._create_input_dict(table=table)

            raw_response = self.client._request(
                method="post",
                path=path,
                query_params=query_params,
                payload=payload,
                file=file,
            )

            return self._parse_table_response(raw_response)

        elif df is not None:
            path = f"/{self.tag}/{name}/upload/json"

            df_json = df.to_dict(orient="records")
            query_params = self._create_input_dict(
                schema=schema, if_exists=if_exists_enum.value
            )
            payload = self._create_input_dict(table=table, data=df_json)

            raw_response = self.client._request(
                method="post",
                path=path,
                query_params=query_params,
                payload=payload,
                file=file,
            )

            return self._parse_table_response(raw_response)

        else:
            raise ThanoSQLValueError("No file or DataFrame provided for upload")

    def delete(self, name: str, schema: Optional[str] = None) -> dict:
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(schema=schema)

        return self.client._request(
            method="delete", path=path, query_params=query_params
        )


class Table(BaseTable):
    service: Optional[TableService] = None

    def get_records(
        self,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> Records:
        path = f"/{self.service.tag}/{self.name}/records"

        query_params = self.service._create_input_dict(
            schema=self.table_schema,
            offset=offset,
            limit=limit,
        )

        res = self.service.client._request(
            method="get",
            path=path,
            query_params=query_params,
        )
        return Records(data=res["records"], total=res["total"])

    def get_records_as_csv(
        self,
        timezone_offset: Optional[int] = None,
    ) -> None:
        path = f"/{self.service.tag}/{self.name}/records/csv"

        query_params = self.service._create_input_dict(
            schema=self.table_schema,
            timezone_offset=timezone_offset,
        )

        self.service.client._request(
            method="get", path=path, query_params=query_params, stream=True
        )

    def insert(
        self,
        records: List[dict],
    ) -> Union[Table, dict]:
        path = f"/{self.service.tag}/{self.name}/records"
        query_params = self.service._create_input_dict(schema=self.table_schema)

        raw_response = self.service.client._request(
            method="post", path=path, query_params=query_params, payload=records
        )

        return self.service._parse_table_response(raw_response)


class TableTemplate(BaseModel):
    name: str
    table_template: TableObject
    version: Optional[str]
    compatibility: Optional[str]
    created_at: Optional[datetime]


class TableTemplateService(ThanoSQLService):
    def __init__(self, client: ThanoSQL) -> None:
        super().__init__(client=client, tag="table_template")

    def list(
        self,
        search: Optional[str] = None,
        order_by: Optional[str] = None,
        latest: Optional[bool] = None,
    ) -> List[TableTemplate]:
        path = f"/{self.tag}/"
        query_params = self._create_input_dict(
            search=search,
            order_by=order_by,
            latest=latest,
        )

        raw_response = self.client._request(
            method="get", path=path, query_params=query_params
        )

        table_templates_adapter = TypeAdapter(List[TableTemplate])
        parsed_response = table_templates_adapter.validate_python(
            raw_response["table_templates"]
        )
        return parsed_response

    def get(self, name: str, version: Optional[str] = None) -> dict:
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(version=version)

        raw_response = self.client._request(
            method="get", path=path, query_params=query_params
        )

        table_templates_adapter = TypeAdapter(List[TableTemplate])
        parsed_response = {}
        parsed_response["table_templates"] = table_templates_adapter.validate_python(
            raw_response["table_templates"]
        )
        parsed_response["versions"] = raw_response["versions"]

        return parsed_response

    def create(
        self,
        name: str,
        table_template: TableObject,
        version: Optional[str] = None,
        compatibility: Optional[str] = None,
    ) -> TableTemplate:
        path = f"/{self.tag}/{name}"
        payload = self._create_input_dict(
            table_template=vars(table_template),
            version=version,
            compatibility=compatibility,
        )

        raw_response = self.client._request(method="post", path=path, payload=payload)

        table_template_adapter = TypeAdapter(TableTemplate)
        parsed_response = table_template_adapter.validate_python(
            raw_response["table_template"]
        )
        return parsed_response

    def delete(self, name: str, version: Optional[str] = None) -> dict:
        path = f"/{self.tag}/{name}"
        query_params = self._create_input_dict(version=version)

        return self.client._request(
            method="delete", path=path, query_params=query_params
        )



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from tests.utils.table import delete_schema
from thanosql._error import ThanoSQLAlreadyExistsError

if TYPE_CHECKING:
    from thanosql._client import ThanoSQL


def test_get_schemas_success(client: ThanoSQL):
    res = client.schema.list()
    assert isinstance(res, dict)
    assert "schemas" in res

    # at least "public" and "qm"
    assert len(res["schemas"]) >= 2


def test_create_schema_same_name(client: ThanoSQL, new_schema: str):
    # schema creation is implicitly tested when calling fixture
    # trying to create a schema of the same name again should result in an error
    with pytest.raises(ThanoSQLAlreadyExistsError):
        client.schema.create(name=new_schema)
